"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const utils_1 = require("@typescript-eslint/utils");
const tsutils_1 = require("tsutils");
const common_1 = require("./common");
/**
 * An ESLint rule to ban partial Array.prototype.reduce().
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noPartialArrayReduce = (0, common_1.createRule)({
    name: "no-partial-array-reduce",
    meta: {
        type: "problem",
        docs: {
            description: "Bans partial Array.prototype.reduce()",
            recommended: "error",
        },
        messages: {
            errorStringGeneric: "Array.prototype.reduce() is partial. It will throw if the array is empty. Provide an initial value or prove the array is non-empty to prevent this error.",
        },
        schema: [],
    },
    create: (context) => {
        const parserServices = utils_1.ESLintUtils.getParserServices(context);
        const checker = parserServices.program.getTypeChecker();
        return {
            // eslint-disable-next-line functional/no-return-void, sonarjs/cognitive-complexity
            CallExpression: (node) => {
                // We only care if this call is a member expression.
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.callee.type !== utils_1.AST_NODE_TYPES.MemberExpression) {
                    return;
                }
                // Non-empty array literal are safe.
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.callee.object.type === utils_1.AST_NODE_TYPES.ArrayExpression &&
                    node.callee.object.elements[0] !== undefined &&
                    node.callee.object.elements[0]?.type !== utils_1.AST_NODE_TYPES.SpreadElement) {
                    return;
                }
                // We only care if this call has exactly one argument.
                // eslint-disable-next-line functional/no-conditional-statements
                if (node.arguments.length !== 1) {
                    return;
                }
                const objectType = checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node.callee.object));
                const typeParts = (0, tsutils_1.unionTypeParts)(objectType);
                // We only care if this call is on an array or tuple (or a type that is a union that includes one or more arrays or tuples)
                // eslint-disable-next-line functional/no-conditional-statements
                if (!typeParts.some((t) => checker.isArrayType(t) ||
                    (0, tsutils_1.isTupleType)(t) ||
                    (0, tsutils_1.isTupleTypeReference)(t))) {
                    return;
                }
                const isProvablyNonEmpty = typeParts.every((t) => {
                    return (((0, tsutils_1.isTupleType)(t) && t.minLength >= 1) ||
                        ((0, tsutils_1.isTupleTypeReference)(t) && t.target.minLength >= 1));
                });
                // eslint-disable-next-line functional/no-conditional-statements
                if (isProvablyNonEmpty) {
                    return;
                }
                const unsafeMethods = [
                    "reduce",
                    "reduceRight",
                ];
                // Detect this form:
                // [].reduce(() => "")
                //
                // Or this form:
                // []["reduce"](() => "")
                //
                // Or this form:
                // const n = "reduce" as const;
                // const foo = [][n](() => "");
                //
                // Or this form:
                // declare const n: "reduce" | "reduceRight";
                // const foo = [""][n](() => "");
                const isReduce = (node.callee.property.type === utils_1.AST_NODE_TYPES.Literal &&
                    node.callee.computed &&
                    typeof node.callee.property.value === "string" &&
                    unsafeMethods.includes(node.callee.property.value)) ||
                    (node.callee.property.type === utils_1.AST_NODE_TYPES.Identifier &&
                        !node.callee.computed &&
                        unsafeMethods.includes(node.callee.property.name)) ||
                    (node.callee.property.type === utils_1.AST_NODE_TYPES.Identifier &&
                        node.callee.computed &&
                        (0, tsutils_1.unionTypeParts)(checker.getTypeAtLocation(parserServices.esTreeNodeToTSNodeMap.get(node.callee.property))).some((type) => type.isStringLiteral() && unsafeMethods.includes(type.value)));
                // eslint-disable-next-line functional/no-conditional-statements
                if (!isReduce) {
                    return;
                }
                // eslint-disable-next-line functional/no-expression-statements
                context.report({
                    node: node,
                    messageId: "errorStringGeneric",
                });
            },
        };
    },
    defaultOptions: [],
});
exports.default = noPartialArrayReduce;
