"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
/* eslint-disable functional/prefer-immutable-types */
const is_immutable_type_1 = require("is-immutable-type");
const tsutils_1 = require("tsutils");
const common_1 = require("./common");
const unsafe_assignment_rule_1 = require("./unsafe-assignment-rule");
/**
 * An ESLint rule to ban unsafe assignment from readonly to mutable types.
 */
// eslint-disable-next-line total-functions/no-unsafe-readonly-mutable-assignment
const noUnsafeReadonlyMutableAssignment = (0, common_1.createRule)({
    name: "no-unsafe-readonly-mutable-assignment",
    meta: {
        type: "problem",
        docs: {
            description: "Bans unsafe assignment from readonly to mutable types.",
            recommended: "error",
        },
        messages: {
            errorStringCallExpression: "Passing a readonly type to a function that expects a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringAssignmentExpression: "Assigning a readonly type to a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringVariableDeclaration: "Using a readonly type to initialize a mutable type can lead to unexpected mutation in the readonly value.",
            errorStringArrowFunctionExpression: "Returning a readonly type from a function that is declared to return a mutable type can lead to unexpected mutation in the readonly value.",
        },
        schema: [],
    },
    create: (0, unsafe_assignment_rule_1.createNoUnsafeAssignmentRule)((checker, destinationType, sourceType) => {
        // eslint-disable-next-line functional/no-conditional-statements
        if (destinationType === sourceType ||
            (0, tsutils_1.unionTypeParts)(destinationType).some((t) => t === sourceType)) {
            // not unsafe
            return false;
        }
        const destinationImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, destinationType);
        const sourceImmutability = (0, is_immutable_type_1.getTypeImmutability)(checker, sourceType);
        return ((destinationImmutability === is_immutable_type_1.Immutability.ReadonlyShallow ||
            destinationImmutability === is_immutable_type_1.Immutability.Mutable) &&
            (sourceImmutability === is_immutable_type_1.Immutability.Immutable ||
                sourceImmutability === is_immutable_type_1.Immutability.ReadonlyDeep));
    }),
    defaultOptions: [],
});
exports.default = noUnsafeReadonlyMutableAssignment;
